from pathlib import Path
from typing import Optional

from truss import __version__

# This needs to be updated whenever we want to update
# base images on a merge. Updating this version will cause
# base images to be pushed with this tag.
TRUSS_BASE_IMAGE_VERSION_TAG = "v0.13.0"


def file_is_empty(path: Path, ignore_hash_style_comments: bool = True) -> bool:
    if not path.exists():
        return True

    with path.open() as file:
        for line in file.readlines():
            if ignore_hash_style_comments and _is_hash_style_comment(line):
                continue
            if line.strip() != "":
                return False

    return True


def file_is_not_empty(path: Path, ignore_hash_style_comments: bool = True) -> bool:
    return not file_is_empty(path, ignore_hash_style_comments)


def _is_hash_style_comment(line: str) -> bool:
    return line.lstrip().startswith("#")


def truss_base_image_tag(
    python_version: str, use_gpu: bool, version_tag: Optional[str] = None
) -> str:
    if version_tag is None:
        version_tag = f"v{__version__}"

    base_tag = python_version
    if use_gpu:
        base_tag = f"{base_tag}-gpu"
    return f"{base_tag}-{version_tag}"



import asyncio
import builtins
import collections
import contextlib
import contextvars
import json
import logging
import statistics
import sys
import textwrap
import threading
import time
import traceback
import typing
from collections.abc import AsyncIterator
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterator,
    Mapping,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

import httpx
import pydantic

from truss.templates.shared import dynamic_config_resolver
from truss_chains import private_types, public_types, utils

if TYPE_CHECKING:
    import aiohttp
    import fastapi

try:
    import prometheus_client
except ImportError:
    logging.warning("Optional `prometheus_client` is not installed.")

    class _NoOpMetric:
        def labels(self, *args: object, **kwargs: object) -> "_NoOpMetric":
            return self

        def inc(self, amount: float = 1.0) -> None:
            pass

        def dec(self, amount: float = 1.0) -> None:
            pass

        def set(self, value: float) -> None:
            pass

        def observe(self, value: float) -> None:
            pass

    class prometheus_client:  # type: ignore[no-redef]
        def Gauge(*args, **kwargs) -> _NoOpMetric:
            return _NoOpMetric()

        def Counter(*args, **kwargs) -> _NoOpMetric:
            return _NoOpMetric()


try:
    from fastapi import WebSocketDisconnect
except ImportError:
    # NB(nikhil): Stub implementation of WebSocketDisconnect, in case local environment doesn't have
    # fastapi.
    class WebSocketDisconnect(Exception):  # type: ignore[no-redef]
        def __init__(self, code: int, reason: Optional[str] = None):
            super().__init__()
            self.code = code
            self.reason = reason


try:
    from starlette.websockets import WebSocketState
except ImportError:
    # NB(nikhil): Stub implementation of WebSocketState, in case local environment doesn't have starlette.
    class WebSocketState:  # type: ignore[no-redef]
        CONNECTED = "connected"


T = TypeVar("T")

_LockT = TypeVar("_LockT", bound=Union[threading.Lock, asyncio.Lock])
_SemaphoreT = TypeVar(
    "_SemaphoreT", bound=Union[threading.Semaphore, asyncio.Semaphore]
)


_ONGOING_REQUESTS = prometheus_client.Gauge(
    name="dependency_chainlet_ongoing_requests",
    documentation="Number of ongoing (executing) requests to dependency Chainlet.",
    labelnames=["dependency_chainlet"],
)

_QUEUED_REQUESTS = prometheus_client.Gauge(
    name="dependency_chainlet_queued_requests",
    documentation="Number of queued (waiting) requests  to dependency Chainlet.",
    labelnames=["dependency_chainlet"],
)
_TOTAL_REQUESTS = prometheus_client.Gauge(
    name="dependency_chainlet_total_requests",
    documentation="Total number of requests (ongoing + queued)  to dependency Chainlet.",
    labelnames=["dependency_chainlet"],
)
_REQUESTS_TOTAL = prometheus_client.Counter(
    name="dependency_chainlet_requests_total",
    documentation="Total number of requests  to dependency Chainlet.",
    labelnames=["dependency_chainlet"],
)


def populate_chainlet_service_predict_urls(
    chainlet_to_service: Mapping[str, private_types.ServiceDescriptor],
) -> Mapping[str, public_types.DeployedServiceDescriptor]:
    chainlet_to_deployed_service: dict[str, public_types.DeployedServiceDescriptor] = {}
    if not chainlet_to_service:
        return {}

    dynamic_chainlet_config_str = dynamic_config_resolver.get_dynamic_config_value_sync(
        private_types.DYNAMIC_CHAINLET_CONFIG_KEY
    )
    if not dynamic_chainlet_config_str:
        raise public_types.MissingDependencyError(
            f"No '{private_types.DYNAMIC_CHAINLET_CONFIG_KEY}' "
            "found. Cannot override Chainlet configs."
        )

    dynamic_chainlet_config = json.loads(dynamic_chainlet_config_str)

    for chainlet_name, service_descriptor in chainlet_to_service.items():
        display_name = service_descriptor.display_name

        # NOTE: The Chainlet `display_name` in the Truss CLI
        # corresponds to Chainlet `name` in the backend. As
        # the dynamic Chainlet config is keyed on the backend
        # Chainlet name, we have to look up config values by
        # using the `display_name` in the service descriptor.
        if display_name not in dynamic_chainlet_config:
            raise public_types.MissingDependencyError(
                f"Chainlet '{display_name}' not found in "
                f"'{private_types.DYNAMIC_CHAINLET_CONFIG_KEY}'. "
                f"Dynamic Chainlet config keys: {list(dynamic_chainlet_config)}."
            )

        if internal_url := dynamic_chainlet_config[display_name].get("internal_url"):
            url = {"internal_url": internal_url}
        else:
            predict_url = dynamic_chainlet_config[display_name].get("predict_url")
            url = {"predict_url": predict_url}

        chainlet_to_deployed_service[chainlet_name] = (
            public_types.DeployedServiceDescriptor(
                display_name=display_name,
                name=service_descriptor.name,
                options=service_descriptor.options,
                **url,
            )
        )

    return chainlet_to_deployed_service


class _BaseSemaphoreWrapper(Generic[_LockT, _SemaphoreT]):
    """Add logging and metrics to semaphore."""

    _lock: _LockT
    _semaphore: _SemaphoreT

    _concurrency_limit: int
    _dependency_chainlet_name: str
    _log_interval_sec: float
    _last_log_time: float
    _pending_count: int
    _wait_times: collections.deque[float]

    def __init__(
        self,
        concurrency_limit: int,
        dependency_chainlet_name: str,
        log_interval_sec: float = 300.0,
    ) -> None:
        self._concurrency_limit = concurrency_limit
        self._dependency_chainlet_name = dependency_chainlet_name
        self._log_interval_sec = log_interval_sec
        self._last_log_time = time.time()
        self._pending_count = 0
        self._wait_times = collections.deque(maxlen=1000)

    @property
    def ongoing_requests(self) -> int:
        return self._concurrency_limit - self._semaphore._value

    @property
    def queued_requests(self) -> int:
        return self._pending_count

    def _maybe_log_stats(self, ongoing_requests: int, queued_requests: int) -> None:
        now = time.time()
        if now - self._last_log_time < self._log_interval_sec:
            return

        self._last_log_time = now

        if not self._wait_times:
            logging.debug(
                f"[{self._dependency_chainlet_name}] no recent requests to log."
            )
            return

        wait_list = list(self._wait_times)
        p50 = statistics.median(wait_list)
        p90 = (
            statistics.quantiles(wait_list, n=10)[8]
            if len(wait_list) >= 10
            else max(wait_list)
        )

        if p50 >= 0.001 or p90 >= 0.001:
            num_waiting = sum(1 for t in wait_list if t > 0.001)
            logging.warning(
                f"Queueing calls to `{self._dependency_chainlet_name}` Chainlet. "
                f"Momentarily there are {ongoing_requests} ongoing requests and "
                f"{queued_requests} waiting requests.\n"
                f"Wait stats: p50={p50:.3f}s, p90={p90:.3f}s.\n"
                f"Of the last {len(wait_list)} requests, {num_waiting} had to wait. "
                f"In many uses cases queueing is fine and does not give a net latency "
                "increase, because the dependency Chainlet replicas cannot process "
                "bursts of requests instantly anyway. Redesigning you algorithm to "
                "send requests more evenly spaced over time could be beneficial and "
                "remove this warning. Alternatively, you could increase "
                f"`concurrency_limit` (currently {self._concurrency_limit}) for "
                f"the {self._dependency_chainlet_name} dependency, but might "
                "risk failures due to overload."
            )
        else:
            logging.debug(
                f"No queueing of calls to `{self._dependency_chainlet_name}`."
            )


class AsyncSemaphoreWrapper(_BaseSemaphoreWrapper[asyncio.Lock, asyncio.Semaphore]):
    def __init__(
        self,
        concurrency_limit: int,
        dependency_chainlet_name: str,
        log_interval_sec: float = 300.0,
    ) -> None:
        super().__init__(concurrency_limit, dependency_chainlet_name, log_interval_sec)
        self._lock = asyncio.Lock()
        self._semaphore = asyncio.Semaphore(concurrency_limit)

    @contextlib.asynccontextmanager
    async def __call__(self) -> AsyncIterator[None]:
        _REQUESTS_TOTAL.labels(dependency_chainlet=self._dependency_chainlet_name).inc()
        _QUEUED_REQUESTS.labels(
            dependency_chainlet=self._dependency_chainlet_name
        ).inc()
        _TOTAL_REQUESTS.labels(dependency_chainlet=self._dependency_chainlet_name).inc()

        start_time = time.perf_counter()

        async with self._lock:
            self._pending_count += 1

        async with self._semaphore:
            wait_duration = time.perf_counter() - start_time
            _QUEUED_REQUESTS.labels(
                dependency_chainlet=self._dependency_chainlet_name
            ).dec()
            _ONGOING_REQUESTS.labels(
                dependency_chainlet=self._dependency_chainlet_name
            ).inc()

            async with self._lock:
                self._pending_count -= 1
                self._wait_times.append(wait_duration)
                self._maybe_log_stats(
                    ongoing_requests=self.ongoing_requests,
                    queued_requests=self.queued_requests,
                )

            try:
                yield
            finally:
                _ONGOING_REQUESTS.labels(
                    dependency_chainlet=self._dependency_chainlet_name
                ).dec()
                _TOTAL_REQUESTS.labels(
                    dependency_chainlet=self._dependency_chainlet_name
                ).dec()


class ThreadSemaphoreWrapper(
    _BaseSemaphoreWrapper[threading.Lock, threading.Semaphore]
):
    def __init__(
        self,
        concurrency_limit: int,
        dependency_chainlet_name: str,
        log_interval_sec: float = 300.0,
    ) -> None:
        super().__init__(concurrency_limit, dependency_chainlet_name, log_interval_sec)
        self._lock = threading.Lock()
        self._semaphore = threading.Semaphore(concurrency_limit)

    @contextlib.contextmanager
    def __call__(self) -> Iterator[None]:
        _REQUESTS_TOTAL.labels(dependency_chainlet=self._dependency_chainlet_name).inc()
        _QUEUED_REQUESTS.labels(
            dependency_chainlet=self._dependency_chainlet_name
        ).inc()
        _TOTAL_REQUESTS.labels(dependency_chainlet=self._dependency_chainlet_name).inc()

        start_time = time.perf_counter()

        with self._lock:
            self._pending_count += 1

        with self._semaphore:
            wait_duration = time.perf_counter() - start_time
            _QUEUED_REQUESTS.labels(
                dependency_chainlet=self._dependency_chainlet_name
            ).dec()
            _ONGOING_REQUESTS.labels(
                dependency_chainlet=self._dependency_chainlet_name
            ).inc()

            with self._lock:
                self._pending_count -= 1
                self._wait_times.append(wait_duration)
                self._maybe_log_stats(
                    ongoing_requests=self.ongoing_requests,
                    queued_requests=self.queued_requests,
                )

            try:
                yield
            finally:
                _ONGOING_REQUESTS.labels(
                    dependency_chainlet=self._dependency_chainlet_name
                ).dec()
                _TOTAL_REQUESTS.labels(
                    dependency_chainlet=self._dependency_chainlet_name
                ).dec()


_trace_parent_context: contextvars.ContextVar[Optional[str]] = contextvars.ContextVar(
    "trace_parent", default=None
)


@contextlib.contextmanager
def _trace_parent(headers: Mapping[str, str]) -> Iterator[None]:
    token = _trace_parent_context.set(
        headers.get(private_types.OTEL_TRACE_PARENT_HEADER_KEY, "")
    )
    try:
        yield
    finally:
        _trace_parent_context.reset(token)


@contextlib.contextmanager
def trace_parent_raw(trace_parent: str) -> Iterator[None]:
    token = _trace_parent_context.set(trace_parent)
    try:
        yield
    finally:
        _trace_parent_context.reset(token)


def get_trace_parent() -> Optional[str]:
    return _trace_parent_context.get()


def pydantic_set_field_dict(obj: pydantic.BaseModel) -> dict[str, pydantic.BaseModel]:
    """Like `BaseModel.model_dump(exclude_unset=True), but only top-level.

    This is used to get kwargs for invoking a function, while dropping fields for which
    there is no value explicitly set in the pydantic model. A field is considered unset
    if the key was not present in the incoming JSON request (from which the model was
    parsed/initialized) and the pydantic model has a default value, such as `None`.

    By dropping these unset fields, the default values from the function definition
    will be used instead. This behavior ensures correct handling of arguments where
    the function has a default, such as in the case of `run_remote`. If the model has
    an optional field defaulting to `None`, this approach differentiates between
    the user explicitly passing a value of `None` and the field being unset in the
    request.

    """
    return {name: getattr(obj, name) for name in obj.model_fields_set}


# Error Propagation Utils. #############################################################


def _handle_exception(exception: Exception) -> NoReturn:
    """Raises `HTTPException` with `RemoteErrorDetail`."""
    try:
        import fastapi
    except ImportError:
        raise utils.make_optional_import_error("fastapi")

    if hasattr(exception, "__module__"):
        exception_module_name = exception.__module__
    else:
        exception_module_name = None

    error_stack = traceback.extract_tb(exception.__traceback__)
    # Filter everything before (model.py) and after (stubs, error handling) so that only
    # user-defined code remains. See test_e2e.py::test_chain for expected results.
    model_predict_index = 0
    first_stub_index = len(error_stack)
    for i, frame in enumerate(error_stack):
        if frame.filename.endswith("model/model.py") and frame.name == "predict":
            model_predict_index = i + 1
        if frame.filename.endswith("remote_chainlet/stub.py") and frame.name.startswith(
            "predict"  # predict sync|async|stream.
        ):
            first_stub_index = i - 1
            break

    final_tb = error_stack[model_predict_index:first_stub_index]
    stack = [
        public_types.RemoteErrorDetail.StackFrame.from_frame_summary(frame)
        for frame in final_tb
    ]
    error = public_types.RemoteErrorDetail(
        exception_cls_name=exception.__class__.__name__,
        exception_module_name=exception_module_name,
        exception_message=str(exception),
        user_stack_trace=list(stack),
    )
    if isinstance(exception, fastapi.HTTPException):
        status_code = exception.status_code
    else:
        status_code = fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR

    raise fastapi.HTTPException(
        status_code=status_code, detail=error.model_dump()
    ) from exception


def _resolve_exception_class(error: public_types.RemoteErrorDetail) -> Type[Exception]:
    """Tries to find the exception class in builtins or imported libs,
    falls back to `public_types.GenericRemoteError` if not found."""
    exception_cls = None
    if error.exception_module_name is None:
        exception_cls = getattr(builtins, error.exception_cls_name, None)
    else:
        if mod := sys.modules.get(error.exception_module_name):
            exception_cls = getattr(mod, error.exception_cls_name, None)

    if exception_cls is None:
        logging.warning(
            f"Could not resolve exception with name `{error.exception_cls_name}` "
            f"and module `{error.exception_module_name}` - fall back to "
            f"`{public_types.GenericRemoteException.__name__}`."
        )
        exception_cls = public_types.GenericRemoteException

    if issubclass(exception_cls, pydantic.ValidationError):
        # Cannot re-raise naively.
        # https://github.com/pydantic/pydantic/issues/6734.
        exception_cls = public_types.GenericRemoteException

    return exception_cls


def _handle_response_error(response_json: dict, base_msg: str, http_status: int):
    try:
        import fastapi
    except ImportError:
        raise utils.make_optional_import_error("fastapi")

    try:
        error_json = response_json["error"]
    except KeyError as e:
        logging.error(f"response_json: {response_json}")
        raise ValueError(
            f"{base_msg}. Could not get `error` field from JSON response."
        ) from e

    try:
        error = public_types.RemoteErrorDetail.model_validate(error_json)
    except pydantic.ValidationError as e:
        if isinstance(error_json, str):
            msg = f"{base_msg}: '{error_json}'"
            raise public_types.GenericRemoteException(msg) from None
        raise ValueError(
            f"{base_msg}: Could not parse chainlet error. Error details are expected "
            "to be either a plain string (old truss models) or a serialized "
            f"`{public_types.RemoteErrorDetail.__name__}`, got:\n{repr(error_json)}"
        ) from e

    exception_cls = _resolve_exception_class(error)
    error_format = textwrap.indent(error.format(), "│   ")
    *lines, last_line = error_format.splitlines()
    last_line = f"╰{last_line[1:]}" if last_line.startswith("│") else last_line
    error_format = "\n".join(lines + [last_line])
    msg = (
        f"(showing chained remote errors, root error at the bottom)\n"
        f"├─ {base_msg}\n"
        f"{error_format}"
    )

    if issubclass(exception_cls, fastapi.HTTPException):
        raise fastapi.HTTPException(status_code=http_status, detail=msg)

    try:
        exc = exception_cls(msg)
    except Exception:  # ruff
        # Some exceptions cannot be directly instantiated with message, fallback.
        exc = public_types.GenericRemoteException(msg)

    raise exc


def _make_base_error_message(remote_name: str, http_status: int) -> str:
    return (
        f"Error calling dependency Chainlet `{remote_name}`, "
        f"HTTP status={http_status}, trace ID=`{get_trace_parent()}`."
    )


def response_raise_errors(response: httpx.Response, remote_name: str) -> None:
    """In case of error, raise it.

    If the response error contains `RemoteErrorDetail`, it tries to re-raise
    the same exception that was raised remotely and falls back to
    `GenericRemoteException` if the exception class could not be resolved.

    Exception messages are chained to trace back to the root cause, i.e. the first
    Chainlet that raised an exception. E.g. the message might look like this:

    ```
    Chainlet-Traceback (most recent call last):
      File "/packages/itest_chain.py", line 132, in run_remote
        value = self._accumulate_parts(text_parts.parts)
      File "/packages/itest_chain.py", line 144, in _accumulate_parts
        value += self._text_to_num.run_remote(part)
    ValueError: (showing chained remote errors, root error at the bottom)
    ├─ Error in dependency Chainlet `TextToNum` (HTTP status 500):
    │   Chainlet-Traceback (most recent call last):
    │     File "/packages/itest_chain.py", line 87, in run_remote
    │       generated_text = self._replicator.run_remote(data)
    │   ValueError: (showing chained remote errors, root error at the bottom)
    │   ├─ Error in dependency Chainlet `TextReplicator` (HTTP status 500):
    │   │   Chainlet-Traceback (most recent call last):
    │   │     File "/packages/itest_chain.py", line 52, in run_remote
    │   │       validate_data(data)
    │   │     File "/packages/itest_chain.py", line 36, in validate_data
    │   │       raise ValueError(f"This input is too long: {len(data)}.")
    ╰   ╰   ValueError: This input is too long: 100.
    ```
    """
    if response.is_error:
        base_msg = _make_base_error_message(remote_name, response.status_code)
        try:
            response_json = response.json()
        except Exception as e:
            raise ValueError(base_msg) from e
        _handle_response_error(response_json, base_msg, response.status_code)


async def async_response_raise_errors(
    response: "aiohttp.ClientResponse", remote_name: str
) -> None:
    """Async version of `async_response_raise_errors`."""
    if response.status >= 400:
        base_msg = _make_base_error_message(remote_name, response.status)
        try:
            response_json = await response.json()
        except Exception as e:
            raise ValueError(base_msg) from e
        _handle_response_error(response_json, base_msg, response.status)


@contextlib.contextmanager
def predict_context(headers: Mapping[str, str]) -> Iterator[None]:
    with _trace_parent(headers):
        try:
            yield
        except Exception as e:
            _handle_exception(e)


class WebsocketWrapperFastAPI:
    """Implements `private_types.WebSocketProtocol` around fastAPI object."""

    # TODO: consider if we want to wrap/translate exceptions thrown as well. Currently
    #  this is somewhat loopy, as `fastapi.WebSocketDisconnect` just passes through,
    #  but we have not documented either, so it's not directly a contradiction.

    def __init__(self, websocket: "fastapi.WebSocket") -> None:
        self._websocket = websocket
        self.headers: Mapping[str, str] = websocket.headers

    async def close(self, code: int = 1000, reason: Optional[str] = None) -> None:
        await self._websocket.close(code=code, reason=reason)

    async def receive(self) -> Union[str, bytes]:
        message = await self._websocket.receive()

        if message.get("type") == "websocket.disconnect":
            # NB(nikhil): Mimics FastAPI `_raise_on_disconnect`, since otherwise the user has no
            # way of detecting that the client disconnected.
            raise WebSocketDisconnect(message["code"], message.get("reason"))
        elif message.get("text"):
            return typing.cast(str, message["text"])
        else:
            return typing.cast(bytes, message["bytes"])

    async def receive_text(self) -> str:
        return await self._websocket.receive_text()

    async def receive_bytes(self) -> bytes:
        return await self._websocket.receive_bytes()

    async def receive_json(self) -> Any:
        return await self._websocket.receive_json()

    async def send_text(self, data: str) -> None:
        await self._websocket.send_text(data)

    async def send_bytes(self, data: bytes) -> None:
        await self._websocket.send_bytes(data)

    async def send_json(self, data: Any) -> None:
        await self._websocket.send_json(data)

    async def iter_text(self) -> AsyncIterator[str]:
        return self._websocket.iter_text()

    async def iter_bytes(self) -> AsyncIterator[bytes]:
        return self._websocket.iter_bytes()

    async def iter_json(self) -> AsyncIterator[Any]:
        return self._websocket.iter_json()

    async def is_connected(self) -> bool:
        # NB(nikhil): This isn't a foolproof mechanism for detecting whether a websocket
        # connection is actually alive, ping/pong messages are best suited for that. However,
        # as a heuristic to determine if a message is safe to send, this can do a pretty good job.
        return (
            self._websocket.application_state == WebSocketState.CONNECTED
            and self._websocket.client_state == WebSocketState.CONNECTED
        )



from truss_train import (
    BasetenCheckpoint,
    CheckpointingConfig,
    Image,
    LoadCheckpointConfig,
    Runtime,
    TrainingJob,
    TrainingProject,
)

load_checkpoint_config = LoadCheckpointConfig(enabled=True)

load_from_most_recent_checkpoint = BasetenCheckpoint.from_latest_checkpoint()

load_most_recent_checkpoint = BasetenCheckpoint.from_latest_checkpoint(
    job_id="lqz4pw4",  # Optional
    project_name="first-project",  # Optional
)

load_from_named_checkpoint = BasetenCheckpoint.from_named_checkpoint(
    checkpoint_name="checkpoint-24", job_id="lqz4pw4"
)

load_checkpoint_config = LoadCheckpointConfig(
    enabled=True,
    download_folder="/tmp/custom_location",  # default is None -> default path set by server-side
    checkpoints=[load_from_most_recent_checkpoint, load_from_named_checkpoint],
)

checkpointing_config = CheckpointingConfig(enabled=True)

job = TrainingJob(
    image=Image(base_image="ghcr.io/baseten-ai/truss-train-base:latest"),
    runtime=Runtime(
        checkpointing_config=checkpointing_config,
        load_checkpoint_config=load_checkpoint_config,
    ),
)

project = TrainingProject(name="new-project", job=job)



import inspect
from collections import Counter

from flask import Flask

# This webserver is meant for testing the patch ping flow in truss
# Some of the end points can be used for patch ping, the names indicate
# the behavior they simulate.
# The stats endpoint can be used to collect stats for verification.

app = Flask(__name__)

_stats = Counter()


def _inc_fn_call_stat():
    global _stats
    fn_name = inspect.stack()[1][3]
    _stats[f"{fn_name}_called_count"] += 1


@app.route("/hash_is_current", methods=["POST"])
def hash_is_current():
    _inc_fn_call_stat()
    return {"is_current": True}


@app.route("/hash_is_current_but_only_every_third_call_succeeds", methods=["POST"])
def hash_is_current_but_only_every_third_call_succeeds():
    _inc_fn_call_stat()
    global _stats
    fn_name = inspect.stack()[0][3]
    if _stats[f"{fn_name}_called_count"] % 3 == 0:
        return {"is_current": True}
    return "simulated failure", 503


@app.route("/accepted", methods=["POST"])
def accepted():
    _inc_fn_call_stat()
    return {"accepted": True}


@app.route("/health")
def health():
    _inc_fn_call_stat()
    return {}


@app.route("/stats")
def stats():
    global _stats
    return _stats



from typing import Optional


def parse_requirement_string(req_str: str) -> Optional[str]:
    """
    Collects requirements from a list of requirement lines.
    """
    stripped_line = req_str.strip()
    if stripped_line and not stripped_line.startswith("#"):
        return stripped_line
    return None


def raise_insufficent_revision(repo_id_huggingface: str, revision: str):
    """
    Raises an exception if the revision is insufficient.
    """
    raise ValueError(
        f"Revision '{revision}' is insufficient for repo '{repo_id_huggingface}'. "
        "Please a suitable commit sha under this "
        f"`[link](https://huggingface.co/{repo_id_huggingface}/commits/main)`"
    )



"""Tests for truss migrate command."""

from truss.base.truss_config import ExternalDataItem, ModelRepo, ModelRepoSourceKind
from truss.cli.migrate_commands import (
    convert_external_data_to_weights,
    convert_model_repo_to_weights,
    generate_mount_location_for_model,
    generate_source_uri,
    migrate_config,
)


class TestGenerateSourceUri:
    """Tests for generate_source_uri function."""

    def test_hf_with_revision(self):
        """HuggingFace repo with revision."""
        model = ModelRepo(
            repo_id="meta-llama/Llama-2-7b",
            revision="main",
            use_volume=True,
            volume_folder="llama",
            kind=ModelRepoSourceKind.HF,
        )
        assert generate_source_uri(model) == "hf://meta-llama/Llama-2-7b@main"

    def test_hf_without_revision(self):
        """HuggingFace repo without revision (v1 style)."""
        model = ModelRepo(
            repo_id="meta-llama/Llama-2-7b",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.HF,
        )
        assert generate_source_uri(model) == "hf://meta-llama/Llama-2-7b"

    def test_gcs_with_prefix(self):
        """GCS with gs:// prefix already."""
        model = ModelRepo(
            repo_id="gs://my-bucket/models",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.GCS,
        )
        assert generate_source_uri(model) == "gs://my-bucket/models"

    def test_gcs_without_prefix(self):
        """GCS without prefix."""
        model = ModelRepo(
            repo_id="my-bucket/models",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.GCS,
        )
        assert generate_source_uri(model) == "gs://my-bucket/models"

    def test_s3_with_prefix(self):
        """S3 with s3:// prefix already."""
        model = ModelRepo(
            repo_id="s3://my-bucket/models",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.S3,
        )
        assert generate_source_uri(model) == "s3://my-bucket/models"

    def test_azure_with_prefix(self):
        """Azure with azure:// prefix already."""
        model = ModelRepo(
            repo_id="azure://account/container",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.AZURE,
        )
        assert generate_source_uri(model) == "azure://account/container"


class TestGenerateMountLocation:
    """Tests for generate_mount_location_for_model function."""

    def test_v2_with_volume_folder(self):
        """v2 model_cache with volume_folder."""
        model = ModelRepo(
            repo_id="meta-llama/Llama-2-7b",
            revision="main",
            use_volume=True,
            volume_folder="llama-7b",
            kind=ModelRepoSourceKind.HF,
        )
        assert generate_mount_location_for_model(model) == "/app/model_cache/llama-7b"

    def test_v1_hf_sanitizes_repo_id(self):
        """v1 HuggingFace sanitizes repo_id (/ -> _)."""
        model = ModelRepo(
            repo_id="meta-llama/Llama-2-7b",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.HF,
        )
        assert (
            generate_mount_location_for_model(model)
            == "/app/model_cache/meta-llama_Llama-2-7b"
        )

    def test_v1_gcs_uses_bucket_name(self):
        """v1 GCS uses bucket name."""
        model = ModelRepo(
            repo_id="gs://my-bucket/path/to/model",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.GCS,
        )
        assert generate_mount_location_for_model(model) == "/app/model_cache/my-bucket"

    def test_v1_s3_uses_bucket_name(self):
        """v1 S3 uses bucket name."""
        model = ModelRepo(
            repo_id="s3://my-s3-bucket/path",
            revision="",
            use_volume=False,
            kind=ModelRepoSourceKind.S3,
        )
        assert (
            generate_mount_location_for_model(model) == "/app/model_cache/my-s3-bucket"
        )


class TestConvertModelRepoToWeights:
    """Tests for convert_model_repo_to_weights function."""

    def test_basic_v2_hf(self):
        """Convert v2 HuggingFace model."""
        model = ModelRepo(
            repo_id="meta-llama/Llama-2-7b",
            revision="main",
            use_volume=True,
            volume_folder="llama-7b",
            kind=ModelRepoSourceKind.HF,
        )
        result = convert_model_repo_to_weights(model)
        assert result["source"] == "hf://meta-llama/Llama-2-7b@main"
        assert result["mount_location"] == "/app/model_cache/llama-7b"
        assert result["auth_secret_name"] == "hf_access_token"

    def test_with_patterns(self):
        """Convert model with allow/ignore patterns."""
        model = ModelRepo(
            repo_id="meta-llama/Llama-2-7b",
            revision="main",
            use_volume=True,
            volume_folder="llama-7b",
            kind=ModelRepoSourceKind.HF,
            allow_patterns=["*.safetensors"],
            ignore_patterns=["*.md"],
        )
        result = convert_model_repo_to_weights(model)
        assert result["allow_patterns"] == ["*.safetensors"]
        assert result["ignore_patterns"] == ["*.md"]

    def test_with_custom_secret(self):
        """Convert model with custom secret name."""
        model = ModelRepo(
            repo_id="meta-llama/Llama-2-7b",
            revision="main",
            use_volume=True,
            volume_folder="llama-7b",
            kind=ModelRepoSourceKind.HF,
            runtime_secret_name="my_hf_token",
        )
        result = convert_model_repo_to_weights(model)
        assert result["auth_secret_name"] == "my_hf_token"


class TestConvertExternalDataToWeights:
    """Tests for convert_external_data_to_weights function."""

    def test_basic_external_data(self):
        """Convert basic external_data item."""
        item = ExternalDataItem(
            url="https://example.com/models/weights.bin",
            local_data_path="weights/model.bin",
        )
        result = convert_external_data_to_weights(item)
        assert result["source"] == "https://example.com/models/weights.bin"
        assert result["mount_location"] == "/app/data/weights/model.bin"

    def test_nested_path(self):
        """Convert external_data with nested path."""
        item = ExternalDataItem(
            url="https://cdn.example.com/deep/nested/path/model.safetensors",
            local_data_path="models/base/adapter.safetensors",
        )
        result = convert_external_data_to_weights(item)
        assert result["mount_location"] == "/app/data/models/base/adapter.safetensors"


class TestMigrateConfig:
    """Tests for migrate_config function."""

    def test_migrate_v2_model_cache(self):
        """Migrate v2 model_cache to weights."""
        config = {
            "model_name": "test",
            "model_cache": [
                {
                    "repo_id": "meta-llama/Llama-2-7b",
                    "revision": "main",
                    "use_volume": True,
                    "volume_folder": "llama-7b",
                }
            ],
        }
        migrated, warnings = migrate_config(config)

        assert "model_cache" not in migrated
        assert "weights" in migrated
        assert len(migrated["weights"]) == 1
        assert migrated["weights"][0]["source"] == "hf://meta-llama/Llama-2-7b@main"
        assert migrated["weights"][0]["mount_location"] == "/app/model_cache/llama-7b"
        assert len(warnings) == 0

    def test_migrate_v1_model_cache_warns(self):
        """Migrate v1 model_cache warns about model.py changes."""
        config = {
            "model_name": "test",
            "model_cache": [{"repo_id": "meta-llama/Llama-2-7b", "use_volume": False}],
        }
        migrated, warnings = migrate_config(config)

        assert "model_cache" not in migrated
        assert "weights" in migrated
        assert len(warnings) == 1
        assert "model.py" in warnings[0]

    def test_migrate_external_data(self):
        """Migrate external_data to weights."""
        config = {
            "model_name": "test",
            "external_data": [
                {
                    "url": "https://example.com/weights.bin",
                    "local_data_path": "weights.bin",
                }
            ],
        }
        migrated, warnings = migrate_config(config)

        assert "external_data" not in migrated
        assert "weights" in migrated
        assert len(migrated["weights"]) == 1
        assert migrated["weights"][0]["source"] == "https://example.com/weights.bin"
        assert migrated["weights"][0]["mount_location"] == "/app/data/weights.bin"

    def test_migrate_combined(self):
        """Migrate both model_cache and external_data."""
        config = {
            "model_name": "test",
            "model_cache": [
                {
                    "repo_id": "meta-llama/Llama-2-7b",
                    "revision": "main",
                    "use_volume": True,
                    "volume_folder": "llama-7b",
                }
            ],
            "external_data": [
                {
                    "url": "https://example.com/adapter.bin",
                    "local_data_path": "adapter.bin",
                }
            ],
        }
        migrated, warnings = migrate_config(config)

        assert "model_cache" not in migrated
        assert "external_data" not in migrated
        assert "weights" in migrated
        assert len(migrated["weights"]) == 2

    def test_migrate_empty_config(self):
        """Empty config returns same config."""
        config = {"model_name": "test"}
        migrated, warnings = migrate_config(config)

        assert migrated == config
        assert len(warnings) == 0

    def test_preserves_other_fields(self):
        """Migration preserves other config fields."""
        config = {
            "model_name": "test",
            "python_version": "py311",
            "requirements": ["torch", "transformers"],
            "model_cache": [
                {
                    "repo_id": "test/model",
                    "revision": "main",
                    "use_volume": True,
                    "volume_folder": "model",
                }
            ],
        }
        migrated, _ = migrate_config(config)

        assert migrated["model_name"] == "test"
        assert migrated["python_version"] == "py311"
        assert migrated["requirements"] == ["torch", "transformers"]



# Import necessary classes from the Baseten Training SDK
from truss_train import definitions
from truss.base import truss_config

PROJECT_NAME = "My-Baseten-Training-Project"
NUM_NODES = 1
NUM_GPUS_PER_NODE = 1

# 1. Define a base image for your training job. You can also use
# private images via AWS IAM or GCP Service Account authentication.
BASE_IMAGE = "pytorch/pytorch:2.7.0-cuda12.8-cudnn9-runtime"

# 2. Define the Runtime Environment for the Training Job
# This includes start commands and environment variables.
# Secrets from the baseten workspace like API keys are referenced using
# `SecretReference`.
training_runtime = definitions.Runtime(
    start_commands=[  # Example: list of commands to run your training script
        "/bin/sh -c 'chmod +x ./run.sh && ./run.sh'"
    ],
    environment_variables={
        # "HF_TOKEN": definitions.SecretReference(name="hf_access_token"),
        "HELLO": "WORLD"
    },
    cache_config=definitions.CacheConfig(
        enabled=False  # Set to True to enable caching between runs
    ),
    checkpointing_config=definitions.CheckpointingConfig(
        enabled=False  # Set to True to enable saving checkpoints on Baseten
    ),
)

training_compute = definitions.Compute(
    node_count=NUM_NODES,
    accelerator=truss_config.AcceleratorSpec(
        accelerator=truss_config.Accelerator.H100, count=NUM_GPUS_PER_NODE
    ),
)

training_job = definitions.TrainingJob(
    image=definitions.Image(base_image=BASE_IMAGE),
    compute=training_compute,
    runtime=training_runtime,
)

training_project = definitions.TrainingProject(name=PROJECT_NAME, job=training_job)



import tempfile
from pathlib import Path

from truss.base.truss_config import DEFAULT_BUNDLED_PACKAGES_DIR
from truss.contexts.local_loader.truss_module_loader import truss_module_loaded

ORIG_MODEL_CLASS_CONTENT = """
class Model:
    x = 1
"""

NEW_MODEL_CLASS_CONTENT = """
class Model:
    x = 2
"""


UTIL_ACCESSOR_MODEL_FILE_CONTENT = """
from model.util import X
from model.submodule.subfile import Y

class Model:
    x = X
    y = Y
"""

UTIL_FILE_CONTENT = """
X = 1
"""

NEW_UTIL_FILE_CONTENT = """
X = 2
"""

SUBMODULE_FILE_CONTENT = """
Y = 3
"""

NEW_SUBMODULE_FILE_CONTENT = """
Y = 4
"""

ORIG_MODEL_CLASS_USING_ADDITIONAL_MODULE_CONTENT = """
from test_module import test
class Model:
    x = test.A
"""

ORIG_ADDITIONAL_MODULE_CONTENT = """
A = 1
"""

NEW_ADDITIONAL_MODULE_CONTENT = """
A = 2
"""


def test_model_module_finder():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(temp_dir, ORIG_MODEL_CLASS_CONTENT)
        with truss_module_loaded(temp_dir, "model.model") as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 1


def test_model_module_finder_reload():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(temp_dir, ORIG_MODEL_CLASS_CONTENT)
        with truss_module_loaded(temp_dir, "model.model") as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 1
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(temp_dir, NEW_MODEL_CLASS_CONTENT)
        with truss_module_loaded(temp_dir, "model.model") as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 2


def test_model_module_finder_additional_modules():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(
            temp_dir,
            ORIG_MODEL_CLASS_USING_ADDITIONAL_MODULE_CONTENT,
            additional_module_file_content=ORIG_ADDITIONAL_MODULE_CONTENT,
        )
        with truss_module_loaded(
            temp_dir, "model.model", DEFAULT_BUNDLED_PACKAGES_DIR
        ) as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 1


def test_model_module_finder_additional_modules_reload():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(
            temp_dir,
            ORIG_MODEL_CLASS_USING_ADDITIONAL_MODULE_CONTENT,
            additional_module_file_content=ORIG_ADDITIONAL_MODULE_CONTENT,
        )
        with truss_module_loaded(
            temp_dir, "model.model", DEFAULT_BUNDLED_PACKAGES_DIR
        ) as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 1

    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(
            temp_dir,
            ORIG_MODEL_CLASS_USING_ADDITIONAL_MODULE_CONTENT,
            additional_module_file_content=NEW_ADDITIONAL_MODULE_CONTENT,
        )
        with truss_module_loaded(
            temp_dir, "model.model", DEFAULT_BUNDLED_PACKAGES_DIR
        ) as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 2


def test_model_module_finder_reload_non_model_file_updated():
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(
            temp_dir,
            UTIL_ACCESSOR_MODEL_FILE_CONTENT,
            UTIL_FILE_CONTENT,
            SUBMODULE_FILE_CONTENT,
        )
        with truss_module_loaded(temp_dir, "model.model") as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 1
            assert model_class.y == 3
    with tempfile.TemporaryDirectory() as temp_dir:
        _write_model_module_files(
            temp_dir,
            UTIL_ACCESSOR_MODEL_FILE_CONTENT,
            NEW_UTIL_FILE_CONTENT,
            NEW_SUBMODULE_FILE_CONTENT,
        )
        with truss_module_loaded(temp_dir, "model.model") as model_module:
            model_class = getattr(model_module, "Model")
            assert model_class.x == 2
            assert model_class.y == 4


def _write_model_module_files(
    truss_dir: str,
    model_file_content: str,
    util_file_content: str = None,
    submodule_file_content: str = None,
    additional_module_file_content: str = None,
):
    model_dir_path = Path(truss_dir) / "model"
    model_dir_path.mkdir(parents=True)
    with (model_dir_path / "model.py").open("w") as f:
        f.write(model_file_content)
    if util_file_content is not None:
        with (model_dir_path / "util.py").open("w") as f:
            f.write(util_file_content)

    if submodule_file_content is not None:
        submodule_path = model_dir_path / "submodule"
        submodule_path.mkdir()
        with (submodule_path / "subfile.py").open("w") as f:
            f.write(submodule_file_content)

    if additional_module_file_content is not None:
        test_module_path = (
            Path(truss_dir) / DEFAULT_BUNDLED_PACKAGES_DIR / "test_module"
        )
        test_module_path.mkdir(parents=True)
        with (test_module_path / "test.py").open("w") as f:
            f.write(additional_module_file_content)
